##############################################################
#
#               GULLIKSEN'S POOL    R   C O D E
#
#       AUTHORS:   URBANO LORENZO-SEVA & PERE J. FERRANDO
#                        URV, TARRAGONA (SPAIN)
#                            DATE: 05/06/2022
#
# DATA EXAMPLE: 61 ITEMS AND N = 578
#
# DATA TEXT FILE USED:    VandoSample1.dat
#
# EXECUTE THE CODE AS:    source("GulliksenPool.r")
#
##############################################################

##############################################################
# UPDATE HERE THE NAME OF YOUR INPUT DATA FILE
# A TEXT FILE WITH NO LABELS IS EXPECTED
#
filein <- "VandoSample1.dat"
# 
# UPDATE HERE THE CODE VALUE FOR MISSING VALUES
# 
MV <- 9
# 
# UPDATE HERE THE NUMBER OF BOOTSTRAP SAMPLES (AT LEAST 500)
# 
K <- 3000
# 
# UPDATE HERE THE CONFIDENCE INTERVAL 
# 
C <- 0.95
#
# 
# UPDATE HERE THE MSA CRITERION TO ELIMINATE ITEMS 
# 
CRIT_MSA <- 0.50
#
#
# UPDATE HERE THE RDI CRITERIA TO INCLUDE ITEMS IN THE POOL (ALWAYS IN LINEAR PARAMETRIZATION)
# 
spool_rdi_inf <- 0.10
spool_rdi_sup <- 0.90
#
#
# UPDATE HERE THE ICI CRITERIA TO INCLUDE ITEMS IN THE POOL  (ALWAYS IN LINEAR PARAMETRIZATION)
# 
spool_ici_inf <- 0.20
spool_ici_sup <- 0.75
# 
#
# UPDATE HERE THE PARAMETRIZATION FOR THE OUTPUT
# 
# VALUE   1   LINEAR (INTERCEPT/CORRELATION)
# VALUE   2   NON LINEAR - ITEM RESPONSE THEORY (THRESHOLD/SLOPE)
#
parametrization <- 2 
#
#
##############################################################

###############  GULLIKSENS POOL CODE #######################
my_invnormal<-function(p) {
#    For small to moderate probabilities, algorithm referenced
#    below is used to obtain an initial approximation which is
#    polished with a final Newton step.
#    For very large arguments, an algorithm of Wichura is used.
#
#  REFERENCE
#
#    Beasley, J. D. and S. G. Springer (1977).
#    Algorithm AS 111: The percentage points of the normal distribution,
#    Applied Statistics, 26, 118-121.
#
#     Wichura, M.J. (1988).
#     Algorithm AS 241: The Percentage Points of the Normal Distribution.
#     Applied Statistics, 37, 477-484.

		if(p < 0 || p > 1) return(NA)		

		if(p == 0)	 return(-Inf)		
		if(p == 1)	 return(Inf)		
		if(p == 0.5) return(0)
    r <- -5
		q <- p - 0.5

		# 0.075 <= p <= 0.925 
    if(abs(q) <= .425) {
			r <- .180625 - q * q
			val <-  q * (((((((r * 2509.0809287301226727 + 33430.575583588128105) * r + 67265.770927008700853) * r + 45921.953931549871457) * r + 13731.693765509461125) * r + 1971.5909503065514427) * r + 133.14166789178437745) * r + 3.387132872796366608) / (((((((r * 5226.495278852854561 + 28729.085735721942674) * r + 39307.89580009271061) * r + 21213.794301586595867) * r + 5394.1960214247511077) * r + 687.1870074920579083) * r + 42.313330701600911252) * r + 1)
		}	else {
   	 	  # closer than 0.075 from {0,1} boundary 
	      # r = min(p, 1-p) < 0.075 
	      if (q > 0) r <- 1 - p
	      else r <- p
        
        r <- sqrt(-log(r))
		    # r = sqrt(-log(r))  <==>  min(p, 1-p) = exp( - r^2 ) 
        
		    if (r <= 5) { # <==> min(p,1-p) >= exp(-25) ~= 1.3888e-11
		    	  r <- r -1.6
		          val <- (((((((r * 7.7454501427834140764e-4 + 0.0227238449892691845833) * r + 0.24178072517745061177) *	r + 1.27045825245236838258) * r + 3.64784832476320460504) * r + 5.7694972214606914055) * r + 4.6303378461565452959) * r + 1.42343711074968357734) / (((((((r * 1.05075007164441684324e-9 + 5.475938084995344946e-4) * r + 0.0151986665636164571966) * r + 0.14810397642748007459) * r + 0.68976733498510000455) * r + 1.6763848301838038494) * r + 2.05319162663775882187) * r + 1)
		      } else { # very close to  0 or 1 
		          r <- r -5
		          val <- (((((((r * 2.01033439929228813265e-7 + 2.71155556874348757815e-5) * r + 0.0012426609473880784386) * r + 0.026532189526576123093) * r + 0.29656057182850489123) * r + 1.7848265399172913358) * r + 5.4637849111641143699) * r + 6.6579046435011037772) / (((((((r * 2.04426310338993978564e-15 + 1.4215117583164458887e-7) * r + 1.8463183175100546818e-5) * r + 7.868691311456132591e-4) * r + 0.0148753612908506148525) * r + 0.13692988092273580531) * r + 0.59983220655588793769) * r + 1)
		      }
		      if (q < 0.0) val <- -val
    }
 return(val)
}

if (K<500) k <- 500
if ((CRIT_MSA < .3) | (CRIT_MSA > .70)) CRIT_MSA <- .5  
X <- read.table(filein)
N <- nrow(X)
m <- ncol(X)

mc <- matrix(0,N,1)
for (i in 1:N){
    mi <- 0
    for (j in 1:m){
        if (X[i,j] == MV){
           mi <- mi+1
           mc[i,1] <- 1
        }
    }
}
Xn <- matrix(0,N-colSums(mc),m)

ni <- 1
for (i in 1:N){
     if (mc[i,1] == 0){
        for (j in 1:m){
           Xn[ni,j] <- X[i,j]
        } 
        ni <- ni +1
    }
}
ni=nrow(Xn)

xmin <- min(Xn)
xmax <- max(Xn)
RDI <- (colSums(Xn)/ni - xmin) / (xmax - xmin) 

R <- cor(Xn)
IR <-solve(R)
SS <-diag(diag(IR)^(-.5))
Q <- SS %*% IR %*% SS
RS <- R * R
QS <- Q * Q
RS <- RS-diag(diag(RS))
QS <- QS-diag(diag(QS))
MNUM <- colSums(RS)
MDEN <- colSums(RS+QS)
MSA <- MNUM/MDEN

ICI <- matrix(0,m)
RE <- matrix(0,m)
deno <- spool_ici_sup * spool_ici_sup * 0.5 * 0.5;
for (j in 1:m){
     ICI[j] <- sqrt(1.0 - 1.0/IR[j,j]) 
     RE[j] <- ICI[j] * ICI[j]*RDI[j]*(1.0 - RDI[j]) / deno;
     if (RE[j] > 1.0) {RE[j]=1.0}
}
  
MSAB <- matrix(0,K,m)
RDIB <- matrix(0,K,m)
ICIB <- matrix(0,K,m)
REB <- matrix(0,K,m)
alive=1
alive2=1
for (i in 1:K){
    if (alive2==10) {
       cat("\014")
       alivestr <- switch(alive,"Computing... \U002B   ","Computing... \U00D7   ")
       print(alivestr)
       alive <- alive+1
       alive2 <- 1
       if (alive>2) alive <- 1
    } else alive2 <- alive2+1

    id <- floor(runif(ni, min=1, max=(ni+1)))
    Xi <- Xn[id,]
    Ri <- cor(Xi)
    IR <-solve(Ri)
    SS <-diag(diag(IR)^(-0.5))
    Q <- SS %*% IR %*% SS
    RS <- Ri * Ri
    QS <- Q * Q
    RS <- RS-diag(diag(RS))
    QS <- QS-diag(diag(QS))
    MNUM <- colSums(RS)
    MDEN <- colSums(RS+QS)
    MSAi <- MNUM/MDEN
    MSAB[i,] <- MSAi
    RDIB[i,] <- (colSums(Xi)/ni - xmin) / (xmax - xmin) 
    
    for (j in 1:m){
       ICIB[i,j] <- sqrt(1.0 - 1.0/IR[j,j]) 
       REB[i,j] <- ICIB[i,j] * ICIB[i,j]*RDIB[i,j]*(1.0 - RDIB[i,j]) / deno;
       if (REB[i,j] > 1.0) {REB[i,j]=1.0}
    }    
}
cat("\014")
print("Job done!   ")
print("            ")

MSAB_IC <- matrix(0,2,m)
RDIB_IC <- matrix(0,2,m)
ICIB_IC <- matrix(0,2,m)
REB_IC <- matrix(0,2,m)
in_mas <- matrix(1,1,m)
in_rdi <- matrix(1,1,m)
in_idi <- matrix(1,1,m)
for (i in 1:m){
    MSAB[,i] = as.matrix(sort(MSAB[,i], method ="quick"))
    MSAB_IC[1,i] = MSAB[floor(K*((1-C)/2)),i]
    MSAB_IC[2,i] = MSAB[floor(K*((1+C)/2)),i]
    in_mas[1,i] <- (MSAB_IC[1,i] > CRIT_MSA) 

    RDIB[,i] = as.matrix(sort(RDIB[,i], method ="quick"))
    RDIB_IC[1,i] = RDIB[floor(K*((1-C)/2)),i]
    RDIB_IC[2,i] = RDIB[floor(K*((1+C)/2)),i]
    in_rdi[1,i] <- ((1-(RDIB_IC[2,i] < spool_rdi_inf)) * (1-(RDIB_IC[1,i] > spool_rdi_sup)))
    
    ICIB[,i] = as.matrix(sort(ICIB[,i], method ="quick"))
    ICIB_IC[1,i] = ICIB[floor(K*((1-C)/2)),i]
    ICIB_IC[2,i] = ICIB[floor(K*((1+C)/2)),i]
    in_idi[1,i] <- ((1-(ICIB_IC[2,i] < spool_ici_inf)) * (1-(ICIB_IC[1,i] > spool_ici_sup)))

    REB[,i] = as.matrix(sort(REB[,i], method ="quick"))
    REB_IC[1,i] = REB[floor(K*((1-C)/2)),i]
    REB_IC[2,i] = REB[floor(K*((1+C)/2)),i]
             
}
eliminate <- m-colSums(t(in_mas*in_rdi*in_idi))

if (parametrization == 2) {
  for (i in 1:m){
    ICI[i] <- ICI[i]/sqrt(1.0 - (ICI[i]*ICI[i])) 
    ICIB_IC[1,i] <- ICIB_IC[1,i]/sqrt(1.0 - (ICIB_IC[1,i]*ICIB_IC[1,i]))  
    ICIB_IC[2,i] <- ICIB_IC[2,i]/sqrt(1.0 - (ICIB_IC[2,i]*ICIB_IC[2,i]))  
   
   RDI[i] <- -1 * my_invnormal(RDI[i])
   RDIB_IC[1,i] <- -1 * my_invnormal(RDIB_IC[1,i])
   RDIB_IC[2,i] <- -1 * my_invnormal(RDIB_IC[2,i])
  
  }   
} 

#################  END OF GULLIKSENS POOL CODE #####################
print("###########################################################################################")
print("#                                                                                         #")
print("#                          GULLIKSEN'S POOL     R   C O D E                               #")
print("#                                                                                         #")
print("#              AUTHORS:   URBANO LORENZO-SEVA & PERE J. FERRANDO                          #")
print("#                              URV, TARRAGONA (SPAIN)                                     #")
print("#                                                                                         #")
print("#                                DATE: 21/09/2022                                         #")
print("#                                                                                         #")
print("###########################################################################################")
print("#                                                                              ")
print(paste("# Filein                : ",filein," "))
print(paste("# Cases                 : ",N," "))
print(paste("# Cases with missings   : ",N-ni," "))
print(paste("# Variables             : ",m," "))
print(paste("# Boostrap samples      : ",K," "))
print(paste("# Percentile            : ",C," "))
if (parametrization == 1) {
   print("# Parametrization       :  LINEAR (INTERCEPT/CORRELATION) ")
} else 
   print("# Parametrization       :  NON LINEAR - ITEM RESPONSE THEORY (THRESHOLD/SLOPE) ")
if (eliminate > 0) {
   print(paste("# Items to revise       : ",eliminate," ")) 
} else 
         print("# Items to revise       : None ")
print("#                                                                                          ")
print("#GULLIKSEN'S POOL: A QUICK TOOL FOR PRELIMINARY DETECTION OF INAPPROPRIATE ITEMS           ")
print("#                                                                                          ")
if (parametrization == 1) {
   print("#STEP 1: The pool based on Relative Difficulty Index (RDI) and Item Consistency Index (ICI)")
} else 
   print("#STEP 1: The pool based on Overall Item Threshold (OIT) and Overall Item Slope (OIS)       ")

print("#                                                                                          ")
print("#                                                                                          ")
if (parametrization == 1) {
   print(sprintf("#Items    RDI     %i%% Confidence         ICI    %i%% Confidence       The Pool                 ",C*100,C*100))
} else 
   print(sprintf("#Items    OIT     %i%% Confidence         OIS    %i%% Confidence       The Pool                 ",C*100,C*100))


print("#                 interval                      interval                                      ")       
print("#-----    ----------------------------   -------------------------   --------                 ")
for (i in 1:m){
   if ((in_rdi[1,i] * in_idi[1,i] )== 0) 
      print(sprintf("# %3i     % 4.3f  (% 4.3f  --  % 4.3f)   %4.3f  (%4.3f  --  %4.3f)   Might not work - Revise  ",i,RDI[i],RDIB_IC[1,i], RDIB_IC[2,i],ICI[i],ICIB_IC[1,i], ICIB_IC[2,i]))
   else 
      print(sprintf("# %3i     % 4.3f  (% 4.3f  --  % 4.3f)   %4.3f  (%4.3f  --  %4.3f)   Admit                    ",i,RDI[i],RDIB_IC[1,i], RDIB_IC[2,i],ICI[i],ICIB_IC[1,i], ICIB_IC[2,i]))
}
print("#-----    ----------------------------   -------------------------   --------                 ")
print("#                                                                                          ")
print("#STEP 2: Assessment of the measure of sampling adequacy (MSA) at the item level            ")
print("#                                                                                          ")
print(sprintf("# Item     MSA    %i%% Confidence       The Pool                                            ",C*100))
print("#                 interval                                                                 ")       
print("# -----    -----  ------------------   ---------                                           ")
for (i in 1:m){
   if (in_rdi[1,i] * in_idi[1,i] > 0) {
     if (in_mas[1,i] == 1) {
        print(sprintf("# %3i      %4.3f  (%4.3f  --  %4.3f)   Admit                                               ",i,MSA[i],MSAB_IC[1,i], MSAB_IC[2,i]))
     } else
        print(sprintf("# %3i      %4.3f  (%4.3f  --  %4.3f)   Might not work - Revise                             ",i,MSA[i],MSAB_IC[1,i], MSAB_IC[2,i]))

   }
}
print("# -----    -----  ------------------   ---------                                           ")

print("#                                                                                          ")
print("#STEP 3: Inspection of the Prior Relative Efficiency for the items admitted in the pool    ")
print("#                                                                                          ")
print(sprintf("# Item     Relative    %i%% Confidence                                                      ",C*100))
print("#          Efficience  interval                                                            ")       
print("# -----    ------------------------------                                                  ")
for (i in 1:m){
   if ((in_mas[1,i] * in_rdi[1,i] * in_idi[1,i] ) > 0) {
        print(sprintf("# %3i      %4.3f       (%4.3f  --  %4.3f)                                                  ",i,RE[i],REB_IC[1,i], REB_IC[2,i]))

   }
}
print("# -----    ------------------------------                                                  ")
print("#                                                                                          ")
print("###########################################################################################")
